import asyncio
import re
from typing import Union
import httpx
import bilix.api.jable as api
from bilix.assign import Handler
from bilix.download.base_downloader_m3u8 import BaseDownloaderM3u8


class DownloaderJable(BaseDownloaderM3u8):
    def __init__(self, videos_dir: str = "videos", video_concurrency: int = 3, part_concurrency: int = 10,
                 speed_limit: Union[float, int] = None, progress=None):
        client = httpx.AsyncClient(
            headers={'user-agent': 'PostmanRuntime/7.29.0', "referer": "https://jable.tv"}, http2=False)
        super(DownloaderJable, self).__init__(client, videos_dir, video_concurrency, part_concurrency,
                                              speed_limit=speed_limit, progress=progress)

    async def get_video(self, url: str, image=True, hierarchy=True):
        video_info = await api.get_video_info(self.client, url)
        if hierarchy:
            hierarchy = self._make_hierarchy_dir(hierarchy, f"{video_info.avid} {video_info.model_name}")
        cors = [self.get_m3u8_video(m3u8_url=video_info.m3u8_url, name=video_info.title,
                                    hierarchy=hierarchy if hierarchy else '')]
        if image:
            cors.append(self._get_static(video_info.img_url, name=video_info.title,
                                         hierarchy=hierarchy if hierarchy else ''))
        await asyncio.gather(*cors)


@Handler('jable')
def handle(**kwargs):
    method = kwargs['method']
    key = kwargs['key']
    videos_dir = kwargs['videos_dir']
    video_concurrency = kwargs['video_concurrency']
    part_concurrency = kwargs['part_concurrency']
    speed_limit = kwargs['speed_limit']
    hierarchy = kwargs['hierarchy']
    image = kwargs['image']
    if 'jable' in key or re.match(r"[A-Za-z]+-\d+", key):
        d = DownloaderJable(videos_dir=videos_dir, video_concurrency=video_concurrency,
                            part_concurrency=part_concurrency, speed_limit=speed_limit)
        if method == 'get_video' or method == 'v':
            cor = d.get_video(key, image=image, hierarchy=hierarchy)
            return d, cor
        raise ValueError(f'For {d.__class__.__name__} "{method}" is not available')



import asyncio
import httpx
from rich.tree import Tree

import bilix.api.bilibili as api
from bilix.log import logger
from bilix.utils import req_retry, convert_size, parse_bilibili_url
from bilix.assign import Handler
from bilix.info.base_informer import BaseInformer

__all__ = ['InformerBilibili']


class InformerBilibili(BaseInformer):
    def __init__(self, sess_data: str = ''):
        cookies = {'SESSDATA': sess_data}
        headers = {'user-agent': 'PostmanRuntime/7.29.0', 'referer': 'https://www.bilibili.com'}
        client = httpx.AsyncClient(headers=headers, cookies=cookies, http2=True)
        super().__init__(client)
        self.type_map = {
            'up': self.info_up,
            'fav': self.info_fav,
            'list': self.info_list,
            'col': self.info_col,
            'video': self.info_video
        }

    async def info_key(self, key):
        t = parse_bilibili_url(key)
        return await self.type_map[t](key)

    async def info_up(self, url: str):
        up_name, total_size, bvids = await api.get_up_info(self.client, url)
        self.console.print(up_name)

    async def info_fav(self, url: str):
        pass

    async def info_list(self, url: str):
        pass

    async def info_col(self, url: str):
        pass

    async def info_video(self, url: str):
        video_info = await api.get_video_info(self.client, url)
        dash = video_info.dash
        if video_info.dash is None:
            logger.warning(f'{video_info.h1_title} 需要大会员或该地区不支持')
            return

        async def ensure_size(data):
            if 'size' not in data:
                res = await req_retry(self.client, data['base_url'], method='GET', headers={'Range': 'bytes=0-1'})
                data['size'] = int(res.headers['Content-Range'].split('/')[-1])

        cors = [ensure_size(d) for d in dash['video']]
        cors.append(ensure_size(dash['audio'][0]))
        if dash['dolby']['audio']:
            cors.append(ensure_size(dash['dolby']['audio'][0]))
        if dash.get('flac', None) and dash['flac']['audio']:  # for some video there is no 'flac' key
            cors.append(ensure_size(dash['flac']['audio']))
        await asyncio.gather(*cors)

        tree = Tree(
            f"[bold reverse] {video_info.h1_title} [/]"
            f" {video_info.status['view']:,}👀 {video_info.status['like']:,}👍 {video_info.status['coin']:,}🪙",
            guide_style="bold cyan")
        video_tree = tree.add("[bold]画面 Video")
        audio_tree = tree.add("[bold]声音 Audio")
        # for video
        for f in video_info.support_formats:
            p_tree = video_tree.add(f['new_description'])
            q_id = f['quality']
            for v in dash['video']:  # todo can be speed up...
                if v['id'] == q_id:
                    p_tree.add(f"codec: {v['codecs']:32} total: {convert_size(v['size'])}")
            if len(p_tree.children) == 0:
                p_tree.style = "rgb(242,93,142)"
                p_tree.add("需要登录或大会员")
        # for audio
        audio_tree.add("默认音质").add(
            f"codec: {dash['audio'][0]['codecs']:32} total: {convert_size(dash['audio'][0]['size'])}")
        if dash['dolby']['type'] != 0:
            sub_tree = audio_tree.add("杜比全景声 Dolby")
            if dash['dolby']['audio']:
                sub_tree.add(
                    f"codec: {dash['dolby']['audio'][0]['codecs']:32}"
                    f" total: {convert_size(dash['dolby']['audio'][0]['size'])}")
            else:
                sub_tree.style = "rgb(242,93,142)"
                sub_tree.add("需要登录或大会员")
        if dash.get('flac', None):
            sub_tree = audio_tree.add("Hi-Res无损")
            if dash['flac']['audio']:
                sub_tree.add(
                    f"codec: {dash['flac']['audio']['codecs']:32}"
                    f" total: {convert_size(dash['flac']['audio']['size'])}")
            else:
                sub_tree.style = "rgb(242,93,142)"
                sub_tree.add("需要登录或大会员")
        self.console.print(tree)


@Handler("bilibili info")
def handle(**kwargs):
    key = kwargs['key']
    method = kwargs['method']
    if 'bilibili' in key and 'info' == method:
        informer = InformerBilibili(sess_data=kwargs['cookie'])
        cor = informer.info_key(key)
        return informer, cor



import threading


class Handler:
    registered = {}
    _lock = threading.RLock()

    def __init__(self, name):
        self.name = name

    def __call__(self, handle_func):
        with self._lock:
            if self.name in self.registered:
                raise KeyError(f"Handler {self.name} all ready exists")
            self.registered[self.name] = handle_func
        return handle_func

    def __repr__(self):
        return f"Handler <name: {self.name} func:{self.registered['name']}>"



from typing import Union, Coroutine, Tuple
from bilix.handle import Handler
from bilix.log import logger
from bilix.download.base_downloader import BaseDownloader
from bilix.info.base_informer import BaseInformer


def assign(**kwargs) -> Tuple[Union[BaseDownloader, BaseInformer], Coroutine]:
    bili_handler = Handler.registered.pop('bilibili')
    for name, handle in Handler.registered.items():
        if res := handle(**kwargs):
            logger.debug(f"Assign to {name}")
            return res
    # since bilix is originally designed for bilibili, finally use bilibili handler
    return bili_handler(**kwargs)



import asyncio
import re
from typing import Union
import aiofiles
import httpx
import os
import m3u8
from Crypto.Cipher import AES
from m3u8 import Segment

from bilix.assign import Handler
from bilix.download.base_downloader import BaseDownloader
from bilix.log import logger
from bilix.utils import req_retry, merge_files


class BaseDownloaderM3u8(BaseDownloader):
    def __init__(self, client: httpx.AsyncClient, videos_dir="videos", video_concurrency=3, part_concurrency=10,
                 speed_limit: Union[float, int] = None, progress=None):
        """
        Base async m3u8 Downloader

        :param client:
        :param videos_dir:
        :param video_concurrency:
        :param part_concurrency:
        :param speed_limit:
        :param progress:
        """
        super(BaseDownloaderM3u8, self).__init__(client, videos_dir, speed_limit=speed_limit, progress=progress)
        self.v_sema = asyncio.Semaphore(video_concurrency)
        self.part_con = part_concurrency
        self.decrypt_cache = {}

    async def _decrypt(self, key: m3u8.Key, content: bytes):
        async def get_key():
            key_bytes = (await req_retry(self.client, uri)).content
            iv = bytes.fromhex(key.iv.replace('0x', ''))
            return AES.new(key_bytes, AES.MODE_CBC, iv)

        uri = key.absolute_uri
        if uri not in self.decrypt_cache:
            self.decrypt_cache[uri] = asyncio.ensure_future(get_key())
            self.decrypt_cache[uri] = await self.decrypt_cache[uri]
        elif asyncio.isfuture(self.decrypt_cache[uri]):
            await self.decrypt_cache[uri]
        cipher = self.decrypt_cache[uri]
        return cipher.decrypt(content)

    async def get_m3u8_video(self, m3u8_url: str, name: str, hierarchy: str = '', retry: int = 5) -> str:
        """
        download

        :param m3u8_url:
        :param name:
        :param hierarchy:
        :param retry:
        :return: downloaded file path
        """
        base_path = f"{self.videos_dir}/{hierarchy}" if hierarchy else self.videos_dir
        file_path = f"{base_path}/{name}.ts"
        if os.path.exists(file_path):
            logger.info(f"[green]已存在[/green] {name}.ts")
            return file_path
        await self.v_sema.acquire()
        res = await req_retry(self.client, m3u8_url)
        m3u8_info = m3u8.loads(res.text)
        if not m3u8_info.base_uri:
            base_uri = re.search(r"(.*)/[^/]+m3u8", m3u8_url).groups()[0]
            m3u8_info.base_uri = base_uri
        cors = []
        p_sema = asyncio.Semaphore(self.part_con)
        task_id = await self.progress.add_task(  # invisible at first and create task id for _get_ts
            description=name if len(name) < 33 else f'{name[:15]}...{name[-15:]}', visible=False)
        total_time = 0
        for idx, seg in enumerate(m3u8_info.segments):
            total_time += seg.duration
            cors.append(self._get_ts(seg, f"{name}-{idx}.ts", task_id, p_sema, hierarchy, retry=retry))
        await self.progress.update(task_id, total=0, total_time=total_time)
        file_list = await asyncio.gather(*cors)
        self.v_sema.release()
        await merge_files(file_list, new_name=file_path)
        logger.info(f"[cyan]已完成[/cyan] {name}.ts")
        await self.progress.update(task_id, visible=False)
        return file_path

    async def _update_task_total(self, task_id, time_part: float, update_size: int):
        task = self.progress.tasks[task_id]
        if not self.progress.tasks[task_id].visible:
            confirmed_t = time_part
            confirmed_b = update_size
            await self.progress.update(task_id, visible=True)
        else:
            confirmed_t = time_part + task.fields['confirmed_t']
            confirmed_b = update_size + task.fields['confirmed_b']
        predicted_total = confirmed_b / confirmed_t * task.fields['total_time']
        await self.progress.update(task_id, total=predicted_total, confirmed_t=confirmed_t, confirmed_b=confirmed_b)

    async def _get_ts(self, seg: Segment, name, task_id, p_sema: asyncio.Semaphore, hierarchy: str = '',
                      retry: int = 5) -> str:
        ts_url = seg.absolute_uri
        base_path = f"{self.videos_dir}/{hierarchy}" if hierarchy else self.videos_dir
        file_path = f"{base_path}/{name}"
        if os.path.exists(file_path):
            downloaded = os.path.getsize(file_path)
            await self._update_task_total(task_id, time_part=seg.duration, update_size=downloaded)
            await self.progress.update(task_id, advance=downloaded)
            return file_path

        async with p_sema:
            content = None
            for exception in range(retry):
                try:
                    content = b''
                    async with self.client.stream("GET", ts_url) as r:
                        r.raise_for_status()
                        await self._update_task_total(
                            task_id, time_part=seg.duration, update_size=int(r.headers['content-length']))
                        async for chunk in r.aiter_bytes():
                            content += chunk
                            await self.progress.update(task_id, advance=len(chunk))
                except (httpx.ReadTimeout, httpx.ConnectTimeout, httpx.RemoteProtocolError) as e:
                    if exception > retry - 2:
                        logger.warning(
                            f'STREAM {e.__class__.__name__} 异常可能由于网络条件不佳或并发数过大导致，若重复出现请考虑降低并发数')
                    await asyncio.sleep(.1)
                except httpx.HTTPStatusError as e:
                    logger.warning(f"STREAM {e}")
                    await asyncio.sleep(.5)
                except Exception as e:
                    if exception > 1:
                        logger.warning(f'STREAM {e.__class__.__name__} 未知异常')
                    await asyncio.sleep(.5)
                else:
                    break
            else:
                logger.error(f"STREAM 超过重复次数 {ts_url}")
                raise Exception(f"STREAM 超过重复次数 {ts_url}")
        # in case .png
        if re.match(r'.*\.ts', ts_url) is None:
            content = content[content.find(b'\x47\x40'):]
        # in case encrypted
        if seg.key:
            content = await self._decrypt(seg.key, content)
        async with aiofiles.open(file_path, 'wb') as f:
            await f.write(content)
        return file_path


@Handler(name="m3u8")
def handle(**kwargs):
    key = kwargs['key']
    if '.m3u8' in key:
        videos_dir = kwargs['videos_dir']
        part_concurrency = kwargs['part_concurrency']
        speed_limit = kwargs['speed_limit']
        method = kwargs['method']
        d = BaseDownloaderM3u8(httpx.AsyncClient(), videos_dir=videos_dir, part_concurrency=part_concurrency,
                               speed_limit=speed_limit)
        if method == 'get_video' or method == 'v':
            cor = d.get_m3u8_video(key, "unnamed")
            return d, cor
        raise ValueError(f'For {d.__class__.__name__} "{method}" is not available')


if __name__ == '__main__':
    async def main():
        async with BaseDownloaderM3u8(httpx.AsyncClient()) as d:
            await d.get_m3u8_video('https://tup.yinghuacd.com/cache/LycorisRecoil07.m3u8',
                                   name="test1")


    asyncio.run(main())



import asyncio
import json
import re
from dataclasses import dataclass
from typing import Union, Sequence
import httpx
import json5

from bilix.dm import parse_view
from bilix.utils import req_retry, legal_title

_dft_headers = {'user-agent': 'PostmanRuntime/7.29.0', 'referer': 'https://www.bilibili.com'}


async def get_cate_meta(client: httpx.AsyncClient) -> dict:
    """
    获取b站分区元数据

    :param client:
    :return:
    """
    cate_info = {}
    res = await req_retry(client, 'https://s1.hdslb.com/bfs/static/laputa-channel/client/assets/index.c0ea30e6.js')
    cate_data = re.search('Za=([^;]*);', res.text).groups()[0]
    cate_data = json5.loads(cate_data)['channelList']
    for i in cate_data:
        if 'sub' in i:
            for j in i['sub']:
                cate_info[j['name']] = j
            cate_info[i['name']] = i
    return cate_info


async def get_list_info(client: httpx.AsyncClient, url_or_sid: str, ):
    """
    获取视频列表信息

    :param url_or_sid:
    :param client:
    :return:
    """
    if url_or_sid.startswith('http'):
        sid = re.search(r'sid=(\d+)', url_or_sid).groups()[0]
    else:
        sid = url_or_sid
    res = await req_retry(client, f'https://api.bilibili.com/x/series/series?series_id={sid}')  # meta api
    meta = json.loads(res.text)
    mid = meta['data']['meta']['mid']
    params = {'mid': mid, 'series_id': sid, 'ps': meta['data']['meta']['total']}
    list_res, up_res = await asyncio.gather(
        req_retry(client, 'https://api.bilibili.com/x/series/archives', params=params),
        req_retry(client, f'https://api.bilibili.com/x/space/acc/info?mid={mid}'))
    list_info, up_info = json.loads(list_res.text), json.loads(up_res.text)
    list_name, up_name = meta['data']['meta']['name'], up_info['data']['name']
    bvids = [i['bvid'] for i in list_info['data']['archives']]
    return list_name, up_name, bvids


async def get_collect_info(client: httpx.AsyncClient, url_or_sid: str):
    """
    获取合集信息

    :param url_or_sid:
    :param client:
    :return:
    """
    sid = re.search(r'sid=(\d+)', url_or_sid).groups()[0] if url_or_sid.startswith('http') else url_or_sid
    params = {'season_id': sid}
    res = await req_retry(client, 'https://api.bilibili.com/x/space/fav/season/list', params=params)
    data = json.loads(res.text)
    medias = data['data']['medias']
    info = data['data']['info']
    col_name, up_name = info['title'], medias[0]['upper']['name']
    bvids = [i['bvid'] for i in data['data']['medias']]
    return col_name, up_name, bvids


async def get_favour_page_info(client: httpx.AsyncClient, url_or_fid: str, pn=1, ps=20, keyword=''):
    """
    获取收藏夹信息（分页）

    :param url_or_fid:
    :param pn:
    :param ps:
    :param keyword:
    :param client:
    :return:
    """
    if url_or_fid.startswith('http'):
        fid = re.findall(r'fid=(\d+)', url_or_fid)[0]
    else:
        fid = url_or_fid
    params = {'media_id': fid, 'pn': pn, 'ps': ps, 'keyword': keyword, 'order': 'mtime'}
    res = await req_retry(client, 'https://api.bilibili.com/x/v3/fav/resource/list', params=params)
    data = json.loads(res.text)['data']
    fav_name, up_name = data['info']['title'], data['info']['upper']['name']
    bvids = [i['bvid'] for i in data['medias'] if i['title'] != '已失效视频']
    total_size = data['info']['media_count']
    return fav_name, up_name, total_size, bvids


async def get_cate_page_info(client: httpx.AsyncClient, cate_id, time_from, time_to, pn=1, ps=30,
                             order='click', keyword=''):
    """
    获取分区视频信息（分页）

    :param cate_id:
    :param pn:
    :param ps:
    :param order:
    :param keyword:
    :param time_from:
    :param time_to:
    :param client:
    :return:
    """
    params = {'search_type': 'video', 'view_type': 'hot_rank', 'cate_id': cate_id, 'pagesize': ps,
              'keyword': keyword, 'page': pn, 'order': order, 'time_from': time_from, 'time_to': time_to}
    res = await req_retry(client, 'https://s.search.bilibili.com/cate/search', params=params)
    info = json.loads(res.text)
    bvids = [i['bvid'] for i in info['result']]
    return bvids


async def get_up_info(client: httpx.AsyncClient, url_or_mid: str, pn=1, ps=30, order='pubdate', keyword=''):
    """
    获取up主信息

    :param url_or_mid:
    :param pn:
    :param ps:
    :param order:
    :param keyword:
    :param client:
    :return:
    """
    if url_or_mid.startswith('http'):
        mid = re.findall(r'/(\d+)', url_or_mid)[0]
    else:
        mid = url_or_mid
    params = {'mid': mid, 'order': order, 'ps': ps, 'pn': pn, 'keyword': keyword}
    res = await req_retry(client, 'https://api.bilibili.com/x/space/arc/search', params=params)
    info = json.loads(res.text)
    up_name = info['data']['list']['vlist'][0]['author']
    total_size = info['data']['page']['count']
    bv_ids = [i['bvid'] for i in info['data']['list']['vlist']]
    return up_name, total_size, bv_ids


@dataclass
class VideoInfo:
    title: str
    h1_title: str
    aid: Union[str, int]
    cid: Union[str, int]
    p: int
    pages: Sequence[Sequence[str]]
    img_url: str
    status: dict
    bvid: str = None
    dash: dict = None
    support_formats: dict = None


async def get_video_info(client: httpx.AsyncClient, url) -> VideoInfo:
    """
    获取视频信息

    :param url:
    :param client:
    :return:
    """
    res = await req_retry(client, url, follow_redirects=True)
    init_info = re.search(r'<script>window.__INITIAL_STATE__=({.*});\(', res.text).groups()[0]  # this line may raise
    init_info = json.loads(init_info)
    if len(init_info.get('error', {})) > 0:
        raise AttributeError("视频已失效")  # 啊叻？视频不见了？在分区下载的时候可能产生
    # extract meta
    pages = []
    h1_title = legal_title(re.search('<h1[^>]*title="([^"]*)"', res.text).groups()[0])
    if 'videoData' in init_info:  # bv视频
        status = {
            'view': init_info['videoData']['stat']['view'],  # 播放量
            'danmaku': init_info['videoData']['stat']['danmaku'],  # 弹幕
            'coin': init_info['videoData']['stat']['coin'],  # 硬币
            'like': init_info['videoData']['stat']['like'],  # 点赞数
            'reply': init_info['videoData']['stat']['reply'],  # 回复数
            'favorite': init_info['videoData']['stat']['favorite'],  # 收藏数
            'share': init_info['videoData']['stat']['share'],  # 分享数
        }
        bvid = init_info['bvid']
        aid = init_info['aid']
        (p, cid), = init_info['cidMap'][bvid]['cids'].items()
        p = int(p) - 1
        title = legal_title(init_info['videoData']['title'])
        base_url = url.split('?')[0]
        for idx, i in enumerate(init_info['videoData']['pages']):
            p_url = f"{base_url}?p={idx + 1}"
            add_name = f"P{idx + 1}-{i['part']}" if len(init_info['videoData']['pages']) > 1 else ''
            pages.append([add_name, p_url])
    elif 'initEpList' in init_info:  # 动漫，电视剧，电影
        status = {
            'view': init_info['mediaInfo']['stat']['views'],  # 播放量
            'danmaku': init_info['mediaInfo']['stat']['danmakus'],  # 弹幕
            'coin': init_info['mediaInfo']['stat']['coins'],  # 硬币
            'like': init_info['mediaInfo']['stat']['likes'],  # 点赞数
            'reply': init_info['mediaInfo']['stat']['reply'],  # 回复数
            'favorite': init_info['mediaInfo']['stat']['favorite'],  # 收藏数
            'favorites': init_info['mediaInfo']['stat']['favorites'],  # 追剧数 / 追番数 （特有）
            'share': init_info['mediaInfo']['stat']['share'],  # 分享数
        }
        bvid = None
        aid = init_info['epInfo']['aid']
        cid = init_info['epInfo']['cid']
        p = init_info['epInfo']['i']
        title = legal_title(re.search('property="og:title" content="([^"]*)"', res.text).groups()[0])
        for idx, i in enumerate(init_info['initEpList']):
            p_url = i['link']
            add_name = i['title']
            pages.append([add_name, p_url])
    else:
        raise AttributeError("未知类型")

    # extract dash
    try:
        play_info = re.search('<script>window.__playinfo__=({.*})</script><script>', res.text).groups()[0]
        play_info = json.loads(play_info)
        dash = play_info['data']['dash']
        support_formats = play_info['data']['support_formats']
    except (KeyError, AttributeError):  # KeyError-电影，AttributeError-动画
        # todo https://www.bilibili.com/video/BV1Jx411r776?p=3 未处理，没有dash下载方式的视频
        dash, support_formats = None, None
    # extract img url
    img_url = re.search('property="og:image" content="([^"]*)"', res.text).groups()[0]
    if not img_url.startswith('http'):  # https://github.com/HFrost0/bilix/issues/52 just for some video
        img_url = 'http:' + img_url.split('@')[0]
    # construct data
    video_info = VideoInfo(title=title, h1_title=h1_title, aid=aid, cid=cid, status=status,
                           p=p, pages=pages, img_url=img_url, bvid=bvid, dash=dash, support_formats=support_formats)
    return video_info


async def get_subtitle_info(client: httpx.AsyncClient, bvid, cid):
    params = {'bvid': bvid, 'cid': cid}
    res = await req_retry(client, 'https://api.bilibili.com/x/player/v2', params=params)
    info = json.loads(res.text)
    if info['code'] == -400:
        raise AttributeError(f'未找到字幕信息')
    return [[f'http:{i["subtitle_url"]}', i['lan_doc']] for i in info['data']['subtitle']['subtitles']]


async def get_dm_info(client: httpx.AsyncClient, aid, cid):
    params = {'oid': cid, 'pid': aid, 'type': 1}
    res = await req_retry(client, f'https://api.bilibili.com/x/v2/dm/web/view', params=params)
    view = parse_view(res.content)
    total = int(view['dmSge']['total'])
    return [f'https://api.bilibili.com/x/v2/dm/web/seg.so?oid={cid}&type=1&segment_index={i + 1}' for i in range(total)]


if __name__ == '__main__':
    import rich

    # result = asyncio.run(get_cate_meta())
    # rich.print(result)
    _dft_client = httpx.AsyncClient(headers=_dft_headers, http2=True)
    result = asyncio.run(get_video_info(
        _dft_client,
        "https://www.bilibili.com/video/BV1fK4y1t7hj"
    ))
    rich.print(result)



import asyncio
import httpx
from typing import Sequence, Union

import bilix.api.yhdmp as api
from bilix.assign import Handler
from bilix.utils import legal_title, cors_slice
from bilix.download.base_downloader_m3u8 import BaseDownloaderM3u8


class DownloaderYhdmp(BaseDownloaderM3u8):
    def __init__(self, videos_dir: str = "videos", video_concurrency: int = 3, part_concurrency: int = 10,
                 speed_limit: Union[float, int] = None, progress=None):
        client = httpx.AsyncClient(
            headers={'user-agent': 'PostmanRuntime/7.29.0', "Referer": "https://www.yhdmp.cc"}, http2=False)
        super(DownloaderYhdmp, self).__init__(client, videos_dir, video_concurrency, part_concurrency,
                                              speed_limit=speed_limit, progress=progress)

    async def get_series(self, url: str, p_range: Sequence[int] = None, hierarchy=True):
        video_info = await api.get_video_info(self.client, url)
        ep_idx = video_info.ep_idx
        play_idx = video_info.play_idx
        title = video_info.title
        if hierarchy:
            hierarchy = self._make_hierarchy_dir(hierarchy, video_info.title)

        # no need to reuse get_video since we only need m3u8_url
        async def get_video(page_url, name):
            m3u8_url = await api.get_m3u8_url(self.client, page_url)
            await self.get_m3u8_video(m3u8_url=m3u8_url, name=name, hierarchy=hierarchy if hierarchy else '')

        cors = []
        for idx, (sub_title, url) in enumerate(video_info.play_info[play_idx]):
            if ep_idx == idx:
                cors.append(self.get_m3u8_video(m3u8_url=video_info.m3u8_url, name=legal_title(title, sub_title),
                                                hierarchy=hierarchy if hierarchy else ''))
            else:
                cors.append(get_video(url, legal_title(title, sub_title)))
        if p_range:
            cors = cors_slice(cors, p_range)
        await asyncio.gather(*cors)

    async def get_video(self, url: str, hierarchy: str = ''):
        video_info = await api.get_video_info(self.client, url)
        name = legal_title(video_info.title, video_info.sub_title)
        await self.get_m3u8_video(m3u8_url=video_info.m3u8_url, name=name, hierarchy=hierarchy)


@Handler(name='樱花动漫P')
def handle(**kwargs):
    method = kwargs['method']
    key = kwargs['key']
    videos_dir = kwargs['videos_dir']
    video_concurrency = kwargs['video_concurrency']
    part_concurrency = kwargs['part_concurrency']
    speed_limit = kwargs['speed_limit']
    hierarchy = kwargs['hierarchy']
    p_range = kwargs['p_range']
    if 'yhdmp' in key:
        d = DownloaderYhdmp(videos_dir=videos_dir, video_concurrency=video_concurrency,
                            part_concurrency=part_concurrency, speed_limit=speed_limit)
        if method == 'get_series' or method == 's':
            cor = d.get_series(key, p_range=p_range, hierarchy=hierarchy)
            return d, cor
        elif method == 'get_video' or method == 'v':
            cor = d.get_video(key)
            return d, cor
        raise ValueError(f'For {d.__class__.__name__} "{method}" is not available')



from typing import Union
import aiofiles
import httpx
import os
from bilix.utils import req_retry
from bilix.log import logger
from bilix.progress import CLIProgress, BaseProgress


class BaseDownloader:
    def __init__(self, client: httpx.AsyncClient, videos_dir='videos', speed_limit: Union[float, int] = None,
                 progress: BaseProgress = None):
        """

        :param client: client used for http request
        :param videos_dir: download to which directory, default to ./videos, if not exists will be auto created
        :param speed_limit: global download rate for the downloader, should be a number (Byte/s unit)
        :param progress: progress obj
        """
        self.client = client
        self.videos_dir = videos_dir
        self.speed_limit = speed_limit
        if not os.path.exists(self.videos_dir):
            os.makedirs(videos_dir)
        if progress is None:
            # if no progress_cls provided by upper class, use cli progress by default
            self.progress = CLIProgress(holder=self)
        else:
            self.progress = progress
            progress.holder = self

    async def __aenter__(self):
        await self.client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.client.__aexit__(exc_type, exc_val, exc_tb)

    async def aclose(self):
        await self.client.aclose()

    def _make_hierarchy_dir(self, hierarchy: Union[bool, str], add_dir: str):
        """
        Make and return new hierarchy according to old hierarchy and add dir

        :param hierarchy: current hierarchy, if True means add_dir becomes new hierarchy
        :param add_dir: new dir add to hierarchy
        :return:
        """
        assert hierarchy is True or (type(hierarchy) is str and len(hierarchy) > 0) and len(add_dir) > 0
        hierarchy = add_dir if hierarchy is True else f'{hierarchy}/{add_dir}'
        if not os.path.exists(f'{self.videos_dir}/{hierarchy}'):
            os.makedirs(f'{self.videos_dir}/{hierarchy}')
        return hierarchy

    async def _get_static(self, url, name, convert_func=None, hierarchy: str = '') -> str:
        """

        :param url:
        :param name: file name (with file type)
        :param convert_func: function used to convert res.content, must be named like ...2...
        :return:
        """
        file_dir = f'{self.videos_dir}/{hierarchy}' if hierarchy else self.videos_dir
        if convert_func:
            file_type = '.' + convert_func.__name__.split('2')[-1]  #
        else:
            file_type = f".{url.split('.')[-1]}" if len(url.split('/')[-1].split('.')) > 1 else ''
            file_type = file_type.split('?')[0]
        file_name = name + file_type
        file_path = f'{file_dir}/{file_name}'
        if os.path.exists(file_path):
            logger.info(f'[green]已存在[/green] {file_name}')  # extra file use different color
        else:
            res = await req_retry(self.client, url)
            content = convert_func(res.content) if convert_func else res.content
            async with aiofiles.open(file_path, 'wb') as f:
                await f.write(content)
            logger.info(f'[cyan]已完成[/cyan] {name + file_type}')  # extra file use different color
        return file_path
